/*
 * @(#)sb16.c  1.1 1999/07/31  Carlos Hasan (chasan@dcc.uchile.cl)
 *
 * Sound Blaster 16 device driver for Intel BeOS Release 4
 *
 * Copyright (C) 1998 Carlos Hasan. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdio.h>
#include <string.h>
#include "sb16.h"
#include "hardware.h"
#include "midi_driver.h"


#define SAMPLES_TO_TIME(samples) ((10000LL * (samples)) / (PCM_SAMPLE_RATE / 100))


/********************************************************************
 * PCM Audio Playback and Capture Constants and Variables
 ********************************************************************/

enum {
    /* playback and capture sampling rate */
    PCM_SAMPLE_RATE = 44100,
	
    /* number of bytes per sample */
    PCM_PLAYBACK_SAMPLE_SIZE = 4,
    PCM_CAPTURE_SAMPLE_SIZE = 2,
	
    /* length of the buffer fragments */
    PCM_PLAYBACK_LATENCY = 2048,
    PCM_CAPTURE_LATENCY = 1024,
	
    /* number of fragments per buffer */
    PCM_PLAYBACK_NUM_BUFFERS = 4,
    PCM_CAPTURE_NUM_BUFFERS = 8,
	
    /* length of the whole DMA buffers */
    PCM_PLAYBACK_BUFFER_SIZE = PCM_PLAYBACK_LATENCY * PCM_PLAYBACK_NUM_BUFFERS,
    PCM_CAPTURE_BUFFER_SIZE = PCM_CAPTURE_LATENCY * PCM_CAPTURE_NUM_BUFFERS,
	
    /* capacity of the sound buffer queues */
    PCM_PLAYBACK_QUEUE_SIZE = 32,
    PCM_CAPTURE_QUEUE_SIZE = 64
};

static struct {
    /* playback and capture semaphores */
    sem_id playback_sem, playback_write, playback_read, playback_wait;
    sem_id capture_sem, capture_write, capture_read, capture_wait;

    /* playback and capture buffer times */	
    bigtime_t playback_time, playback_clock;
    bigtime_t capture_time, capture_clock;

	/* playback and capture DMA buffers */	
    hw_dma_buffer playback_buffer, capture_buffer;
    int playback_offset, playback_underflow;
    int capture_offset, capture_overflow;
	
    /* playback and capture sound buffer queues */
    sound_buffer_header *playback_queue[PCM_PLAYBACK_QUEUE_SIZE];
    sound_buffer_header *capture_queue[PCM_CAPTURE_QUEUE_SIZE];	
    int playback_queue_read, playback_queue_write;
    int capture_queue_read, capture_queue_write;

	/* playback and capture threads */
    bool playback_thread_active, capture_thread_active;
    thread_id playback_thread, capture_thread;
} pcm;

/********************************************************************
 * PCM Audio Playback Routines
 ********************************************************************/

static void pcm_playback_callback(void)
{
    /* update the playback interrupt time and sample clock */
    pcm.playback_time = system_time();
    pcm.playback_clock += PCM_PLAYBACK_LATENCY;

    /* wake up the playback thread when there is work to do */
    release_sem_etc(pcm.playback_wait, 1, B_DO_NOT_RESCHEDULE);
}

static int32 pcm_playback_thread(void *args)
{
    sound_buffer_header *buffer_header;
    int buffer_length, to_write, written;
    char *buffer_ptr;
    bigtime_t delta_time;
	
    /* start the playback DMA activity */
    hw_clear_playback_interrupts();
    hw_start_playback();

    /* start with an empty sound buffer */
    buffer_ptr = 0;
    buffer_length = 0;

    while (pcm.playback_thread_active) {

	/* wait the next DMA playback interrupt */
	if (acquire_sem_etc(pcm.playback_wait, 1, B_RELATIVE_TIMEOUT, 1000000LL) != B_NO_ERROR)
	    break;

	/* fill the next fragment of the playback buffer */
	to_write = PCM_PLAYBACK_LATENCY;
	while (to_write > 0) {
			
	    /* try to get a sound buffer when the current buffer is empty */
	    if (buffer_length <= 0) {
			if (acquire_sem_etc(pcm.playback_read, 1, B_RELATIVE_TIMEOUT, 0) == B_NO_ERROR) {
			    /* get the next sound buffer from the head of the playback queue */
			    buffer_header = pcm.playback_queue[pcm.playback_queue_read++];
			    if (pcm.playback_queue_read >= PCM_PLAYBACK_QUEUE_SIZE)
				pcm.playback_queue_read = 0;
			    release_sem_etc(pcm.playback_write, 1, B_DO_NOT_RESCHEDULE);
						
			    /* set the playback time and sample clock of the sound buffer */
			    delta_time = SAMPLES_TO_TIME((to_write - PCM_PLAYBACK_LATENCY) / PCM_PLAYBACK_SAMPLE_SIZE);
			    buffer_header->time = pcm.playback_time - delta_time;
			    buffer_header->sample_clock = SAMPLES_TO_TIME(pcm.playback_clock / PCM_PLAYBACK_SAMPLE_SIZE) - delta_time;
						
			    /* get the data address and length of the sound buffer */
			    buffer_ptr = (char *) buffer_header + sizeof(sound_buffer_header);
			    buffer_length = buffer_header->reserved_1 - sizeof(sound_buffer_header);
			}
	    }

	    written = to_write;
	    if (buffer_length <= 0) {
			/* fill with silence the playback buffer fragment when the sound buffer is empty */
			memset(pcm.playback_buffer.address + pcm.playback_offset, 0, written);
			pcm.playback_underflow += written;
	    }
	    else {
			/* reset the underflow counter of the playback buffer */
			pcm.playback_underflow = 0;
				
			/* copy the data from the sound buffer to the playback buffer */
			if (written > buffer_length)
			    written = buffer_length;
			memcpy(pcm.playback_buffer.address + pcm.playback_offset, buffer_ptr, written);
				
			/* release the sound buffer when it is full */
			buffer_ptr += written;
			if ((buffer_length -= written) <= 0)
			    release_sem_etc(pcm.playback_sem, 1, B_DO_NOT_RESCHEDULE);
		    }
	
		    /* advance the write offset of the playback buffer */
		    if ((pcm.playback_offset += written) >= PCM_PLAYBACK_BUFFER_SIZE)
			pcm.playback_offset = 0;
		    to_write -= written;
		}
		
		/* suspend the playback thread when the buffer is filled with silence */
		if (pcm.playback_underflow >= PCM_PLAYBACK_BUFFER_SIZE) {
		    hw_pause_playback();
		    suspend_thread(pcm.playback_thread);
		    hw_resume_playback();
		}
    }
	
    /* stop the DMA playback activity */
    hw_stop_playback();
    hw_clear_playback_interrupts();
		
    return 0;
}


/********************************************************************
 * PCM Audio Capture Routines
 ********************************************************************/

static void pcm_capture_callback(void)
{
    /* update the capture interrupt time and sample clock */
    pcm.capture_time = system_time();
    pcm.capture_clock += PCM_CAPTURE_LATENCY;

    /* wake up the capture thread when there is work to do */
    release_sem_etc(pcm.capture_wait, 1, B_DO_NOT_RESCHEDULE);
}

static int32 pcm_capture_thread(void *args)
{
    sound_buffer_header *buffer_header;
    int buffer_length, to_write, written, index;
    short *buffer_ptr;
    bigtime_t delta_time;
	
    /* start the capture DMA activity */
    hw_clear_capture_interrupts();
    hw_start_capture();

    /* start with an empty sound buffer */
    buffer_ptr = 0;
    buffer_length = 0;

    while (pcm.capture_thread_active) {

		/* wait the next DMA capture interrupt */
		if (acquire_sem_etc(pcm.capture_wait, 1, B_RELATIVE_TIMEOUT, 1000000LL) != B_NO_ERROR)
		    break;
	
		/* copy the next fragment from the capture buffer */
		to_write = PCM_CAPTURE_LATENCY;
		while (to_write > 0) {
				
		    /* try to get a sound buffer when the current buffer is full */
		    if (buffer_length <= 0) {
				if (acquire_sem_etc(pcm.capture_read, 1, B_RELATIVE_TIMEOUT, 0) == B_NO_ERROR) {
				    /* get the next sound buffer from the head of the capture queue */
				    buffer_header = pcm.capture_queue[pcm.capture_queue_read++];
				    if (pcm.capture_queue_read >= PCM_CAPTURE_QUEUE_SIZE)
					pcm.capture_queue_read = 0;
				    release_sem_etc(pcm.capture_write, 1, B_DO_NOT_RESCHEDULE);
							
				    /* set the capture time and sample clock of the sound buffer */
				    delta_time = SAMPLES_TO_TIME(to_write / PCM_CAPTURE_SAMPLE_SIZE);
				    buffer_header->time = pcm.capture_time - delta_time;
				    buffer_header->sample_clock = SAMPLES_TO_TIME(pcm.capture_clock / PCM_CAPTURE_SAMPLE_SIZE) - delta_time;
							
				    /* get the data address and length of the sound buffer */
				    buffer_ptr = (short *) ((char *) buffer_header + sizeof(sound_buffer_header));
				    buffer_length = (buffer_header->reserved_1 - sizeof(sound_buffer_header)) >> 1;
				}
		    }
	
		    written = to_write;
		    if (buffer_length <= 0) {
				/* drop the capture buffer fragment when the sound buffer is full */
				pcm.capture_overflow += written;
		    }
		    else {
				/* reset the overflow counter of the capture buffer */
				pcm.capture_overflow = 0;
					
				/* copy the data from the capture buffer to the sound buffer (8-bit to 16-bit PCM linear) */
				if (written > buffer_length)
				    written = buffer_length;
				for (index = 0; index < written; index++)
				    buffer_ptr[index] = (pcm.capture_buffer.address[pcm.capture_offset + index] ^ 0x80) << 8;
					
				/* release the sound buffer when it is empty */
				buffer_ptr += written;
				if ((buffer_length -= written) <= 0)
				    release_sem_etc(pcm.capture_sem, 1, B_DO_NOT_RESCHEDULE);
		    }
		
		    /* advance the read offset of the capture buffer */
		    if ((pcm.capture_offset += written) >= PCM_CAPTURE_BUFFER_SIZE)
			pcm.capture_offset = 0;
		    to_write -= written;
		}
			
		/* suspend the capture thread when a while buffer has been ignored */
		if (pcm.capture_overflow >= PCM_CAPTURE_BUFFER_SIZE) {
		    hw_pause_capture();
		    suspend_thread(pcm.capture_thread);
		    hw_resume_capture();
		}
    }
	
    /* stop the DMA capture activity */
    hw_stop_capture();
    hw_clear_capture_interrupts();
	
    return 0;
}


/********************************************************************
 * PCM Audio Initialization and Control Routines
 ********************************************************************/

static status_t pcm_init(void)
{
    memset(&pcm, 0, sizeof(pcm));

    /* initialize the playback and capture semaphores */
    pcm.playback_write = create_sem(PCM_PLAYBACK_QUEUE_SIZE, "sb16 playback write");
    pcm.capture_write = create_sem(PCM_CAPTURE_QUEUE_SIZE, "sb16 capture write");

    pcm.playback_read = create_sem(0, "sb16 playback read");
    pcm.capture_read = create_sem(0, "sb16 capture read");

    pcm.playback_wait = create_sem(0, "sb16 playback wait");
    pcm.capture_wait = create_sem(0, "sb16 capture wait");

    set_sem_owner(pcm.playback_write, B_SYSTEM_TEAM);
    set_sem_owner(pcm.capture_write, B_SYSTEM_TEAM);
	
    set_sem_owner(pcm.playback_read, B_SYSTEM_TEAM);
    set_sem_owner(pcm.capture_read, B_SYSTEM_TEAM);
	
    set_sem_owner(pcm.playback_wait, B_SYSTEM_TEAM);
    set_sem_owner(pcm.capture_wait, B_SYSTEM_TEAM);
	
    /* initialize the playback and capture time stamps */
    pcm.playback_time = pcm.capture_time = system_time();
    pcm.playback_clock = pcm.capture_clock = 0;

    /* allocate the playback and capture DMA buffers */	
    hw_alloc_buffer(&pcm.playback_buffer, 2 * PCM_PLAYBACK_BUFFER_SIZE);
    hw_alloc_buffer(&pcm.capture_buffer, 2 * PCM_CAPTURE_BUFFER_SIZE);
	
    /* adjust the size of the allocated DMA buffers */
    pcm.playback_buffer.size = PCM_PLAYBACK_BUFFER_SIZE;
    pcm.capture_buffer.size = PCM_CAPTURE_BUFFER_SIZE;
	
    memset(pcm.playback_buffer.address, 0, PCM_PLAYBACK_BUFFER_SIZE);
    memset(pcm.capture_buffer.address, 0, PCM_CAPTURE_BUFFER_SIZE);
	
    hw_set_playback_buffer(&pcm.playback_buffer, 16, 2, PCM_PLAYBACK_LATENCY);
    hw_set_capture_buffer(&pcm.capture_buffer, 8, 2, PCM_CAPTURE_LATENCY);	
		
    pcm.playback_offset = pcm.playback_underflow = 2 * PCM_PLAYBACK_LATENCY; 
    pcm.capture_offset = pcm.capture_overflow = 0;

    /* initialize the playback and capture sampling rates */
    hw_set_playback_rate(PCM_SAMPLE_RATE);
    hw_set_capture_rate(PCM_SAMPLE_RATE);

    /* initialize the playback and capture queue */	
    pcm.playback_queue_read = pcm.playback_queue_write = 0;
    pcm.capture_queue_read = pcm.capture_queue_write = 0;

    /* set up the playback and capture callbacks */
    hw_set_playback_callback(pcm_playback_callback);
    hw_set_capture_callback(pcm_capture_callback);

    /* enable playback and capture threads */
    pcm.playback_thread_active = true;
    pcm.capture_thread_active = true;

    /* create the playback and capture threads */
    pcm.playback_thread = spawn_kernel_thread(pcm_playback_thread, "sb16 playback thread", B_REAL_TIME_PRIORITY, 0);
    pcm.capture_thread = spawn_kernel_thread(pcm_capture_thread, "sb16 capture thread", B_REAL_TIME_PRIORITY, 0);
	
    /* resume the playback and capture threads */
    resume_thread(pcm.playback_thread);
    resume_thread(pcm.capture_thread);
	
    return B_NO_ERROR;
}

static status_t pcm_uninit(void)
{
    status_t rc;

    /* disable the playback and capture threads */
    pcm.playback_thread_active = false;
    pcm.capture_thread_active = false;

    /* resume and wait until the threads are killed */
    wait_for_thread(pcm.playback_thread, &rc);
    wait_for_thread(pcm.capture_thread, &rc);

    /* release the playback and capture callbacks */
    hw_set_playback_callback(0);
    hw_set_capture_callback(0);
		
    /* release the playback and capture buffers */
    hw_free_buffer(&pcm.capture_buffer);
    hw_free_buffer(&pcm.playback_buffer);

    /* release the playback and capture semaphores */
    delete_sem(pcm.capture_wait);
    delete_sem(pcm.playback_wait);
		
    delete_sem(pcm.capture_read);
    delete_sem(pcm.playback_read);

    delete_sem(pcm.capture_write);
    delete_sem(pcm.playback_write);
	
    return B_NO_ERROR;
}

static status_t pcm_get_params(sound_params *params)
{
    int mux;
	
    /* hardcoded sampling rate and formats */
    params->sample_rate = SOUND_44100_HZ;
    params->playback_format = SOUND_PCM_STEREO_16_BIT;
    params->capture_format = SOUND_PCM_STEREO_16_BIT;
	
    /* get the hardware mixer settings */
    params->dither_enable = 0;
    params->mic_attn = 63 - (hw_get_mixer_value(SB16_MIC) >> 2);
    params->mic_enable = (hw_get_mixer_value(SB16_OUTPUT_MUX) & B_SB16_MUX_MIC) != 0;
    params->mono_gain = 63 - (hw_get_mixer_value(SB16_PC_BEEP) >> 2);
    params->mono_mute = 0;
	
    mux = hw_get_mixer_value(SB16_INPUT_MUX_LEFT);	
    params->left.adc_source = 
		(mux & B_SB16_MUX_LINE_LEFT) ? SOUND_ADC_LINE :
		(mux & B_SB16_MUX_CD_LEFT) ? SOUND_ADC_CD :
		(mux & B_SB16_MUX_MIC) ? SOUND_ADC_MIC : SOUND_ADC_LOOPBACK;
    params->left.adc_gain = hw_get_mixer_value(SB16_ADC_GAIN_LEFT) >> 4;
    params->left.mic_gain_enable = hw_get_mixer_value(SB16_MIC_BOOST) != 0;
    params->left.cd_mix_gain = 31 - (hw_get_mixer_value(SB16_CD_LEFT) >> 3);
    params->left.cd_mix_mute = (hw_get_mixer_value(SB16_OUTPUT_MUX) & B_SB16_MUX_CD_LEFT) == 0;
    params->left.line_mix_gain = 31 - (hw_get_mixer_value(SB16_LINE_IN_LEFT) >> 3);
    params->left.line_mix_mute = (hw_get_mixer_value(SB16_OUTPUT_MUX) & B_SB16_MUX_LINE_LEFT) == 0;
    params->left.dac_attn = 63 - (hw_get_mixer_value(SB16_DAC_LEFT) >> 2);
    params->left.dac_mute = (hw_get_mixer_value(SB16_LINE_OUT_LEFT) == 0);
    params->left.reserved_1 = (hw_get_mixer_value(SB16_TREBLE_LEFT) >> 2) - 32;
    params->left.reserved_2 = (hw_get_mixer_value(SB16_BASS_LEFT) >> 2) - 32;
	
    mux = hw_get_mixer_value(SB16_INPUT_MUX_RIGHT);
    params->right.adc_source = 
		(mux & B_SB16_MUX_LINE_RIGHT) ? SOUND_ADC_LINE :
		(mux & B_SB16_MUX_CD_RIGHT) ? SOUND_ADC_CD :
		(mux & B_SB16_MUX_MIC) ? SOUND_ADC_MIC : SOUND_ADC_LOOPBACK;
    params->right.adc_gain = hw_get_mixer_value(SB16_ADC_GAIN_RIGHT) >> 4;
    params->right.mic_gain_enable = hw_get_mixer_value(SB16_MIC_BOOST) != 0;
    params->right.cd_mix_gain = 31 - (hw_get_mixer_value(SB16_CD_RIGHT) >> 3);
    params->right.cd_mix_mute = (hw_get_mixer_value(SB16_OUTPUT_MUX) & B_SB16_MUX_CD_RIGHT) == 0;
    params->right.line_mix_gain = 31 - (hw_get_mixer_value(SB16_LINE_IN_RIGHT) >> 3);
    params->right.line_mix_mute = (hw_get_mixer_value(SB16_OUTPUT_MUX) & B_SB16_MUX_LINE_RIGHT) == 0;
    params->right.dac_attn = 63 - (hw_get_mixer_value(SB16_DAC_RIGHT) >> 2);
    params->right.dac_mute = (hw_get_mixer_value(SB16_LINE_OUT_RIGHT) == 0);
    params->right.reserved_1 = (hw_get_mixer_value(SB16_TREBLE_RIGHT) >> 2) - 32;
    params->right.reserved_2 = (hw_get_mixer_value(SB16_BASS_RIGHT) >> 2) - 32;

    return B_NO_ERROR;
}

static status_t pcm_set_params(sound_params *params)
{
    int mux;
	
    /* set the hardware mixer settings */
    hw_set_mixer_value(SB16_MIC, (63 - params->mic_attn) << 3);
    hw_set_mixer_value(SB16_PC_BEEP, (63 - params->mono_gain) << 2);

    mux = 
		(params->left.adc_source == SOUND_ADC_LINE) ? B_SB16_MUX_LINE_LEFT :
		(params->left.adc_source == SOUND_ADC_CD) ? B_SB16_MUX_CD_LEFT :
		(params->left.adc_source == SOUND_ADC_MIC) ? B_SB16_MUX_MIC : 0;
    hw_set_mixer_value(SB16_INPUT_MUX_LEFT, mux);
    hw_set_mixer_value(SB16_ADC_GAIN_LEFT, params->left.adc_gain << 4);
    hw_set_mixer_value(SB16_CD_LEFT, (31 - params->left.cd_mix_gain) << 3);
    hw_set_mixer_value(SB16_LINE_IN_LEFT, (31 - params->left.line_mix_gain) << 3);
    hw_set_mixer_value(SB16_DAC_LEFT, (63 - params->left.dac_attn) << 2);
    hw_set_mixer_value(SB16_LINE_OUT_LEFT, params->left.dac_mute ? 0 : 248);	
    hw_set_mixer_value(SB16_TREBLE_LEFT, (32 + params->left.reserved_1) << 2);
    hw_set_mixer_value(SB16_BASS_LEFT, (32 + params->left.reserved_2) << 2);
	
    mux = 
		(params->right.adc_source == SOUND_ADC_LINE) ? B_SB16_MUX_LINE_RIGHT :
		(params->right.adc_source == SOUND_ADC_CD) ? B_SB16_MUX_CD_RIGHT :
		(params->right.adc_source == SOUND_ADC_MIC) ? B_SB16_MUX_MIC : 0;
    hw_set_mixer_value(SB16_INPUT_MUX_RIGHT, mux);
    hw_set_mixer_value(SB16_ADC_GAIN_RIGHT, params->right.adc_gain << 4);
    hw_set_mixer_value(SB16_CD_RIGHT, (31 - params->right.cd_mix_gain) << 3);
    hw_set_mixer_value(SB16_LINE_IN_RIGHT, (31 - params->right.line_mix_gain) << 3);
    hw_set_mixer_value(SB16_DAC_RIGHT, (63 - params->right.dac_attn) << 2);
    hw_set_mixer_value(SB16_LINE_OUT_RIGHT, params->right.dac_mute ? 0 : 248);
    hw_set_mixer_value(SB16_TREBLE_RIGHT, (32 + params->right.reserved_1) << 2);
    hw_set_mixer_value(SB16_BASS_RIGHT, (32 + params->right.reserved_2) << 2);

    mux =
		(params->mic_enable ? B_SB16_MUX_MIC : 0) |
		(params->left.cd_mix_mute ? 0 : B_SB16_MUX_CD_LEFT) |
		(params->left.line_mix_mute ? 0 : B_SB16_MUX_LINE_LEFT) |
		(params->right.cd_mix_mute ? 0 : B_SB16_MUX_CD_RIGHT) |
		(params->right.line_mix_mute ? 0 : B_SB16_MUX_LINE_RIGHT);
    hw_set_mixer_value(SB16_OUTPUT_MUX, mux);
    hw_set_mixer_value(SB16_MIC_BOOST, params->left.mic_gain_enable | params->right.mic_gain_enable);	

    return B_NO_ERROR;
}

static status_t pcm_set_playback_sem(sem_id sem)
{
    pcm.playback_sem = sem;
    return B_NO_ERROR;
}

static status_t pcm_set_capture_sem(sem_id sem)
{
    pcm.capture_sem = sem;
    return B_NO_ERROR;
}

static status_t pcm_write_buffer(sound_buffer_header *header)
{
    /* resume the playback thread when there is work to do */
    if (pcm.playback_underflow >= PCM_PLAYBACK_BUFFER_SIZE)
		resume_thread(pcm.playback_thread);
	
    if (acquire_sem_etc(pcm.playback_write, 1, B_RELATIVE_TIMEOUT, 1000000LL) == B_NO_ERROR) {
		/* put a new sound buffer in the playback queue */
		pcm.playback_queue[pcm.playback_queue_write++] = header;
		if (pcm.playback_queue_write >= PCM_PLAYBACK_QUEUE_SIZE)
		    pcm.playback_queue_write = 0;
		release_sem_etc(pcm.playback_read, 1, B_DO_NOT_RESCHEDULE);
	
		return B_NO_ERROR;
    }

    return B_ERROR;
}

static status_t pcm_read_buffer(sound_buffer_header *header)
{
    /* resume the capture thread when there is work to do */
    if (pcm.capture_overflow >= PCM_CAPTURE_BUFFER_SIZE)
		resume_thread(pcm.capture_thread);
	
    if (acquire_sem_etc(pcm.capture_write, 1, B_RELATIVE_TIMEOUT, 1000000LL) == B_NO_ERROR) {
		/* put a new sound buffer in the capture queue */
		pcm.capture_queue[pcm.capture_queue_write++] = header;
		if (pcm.capture_queue_write >= PCM_CAPTURE_QUEUE_SIZE)
		    pcm.capture_queue_write = 0;
		release_sem_etc(pcm.capture_read, 1, B_DO_NOT_RESCHEDULE);
	
		return B_NO_ERROR;
    }

    return B_ERROR;
}

static status_t pcm_lock_for_dma(void)
{
    return B_NO_ERROR;
}


/********************************************************************
 * PCM Audio Driver External Interface
 ********************************************************************/

static uint32 pcm_open_count = 0, midi_open_count = 0;

static status_t pcm_open(const char *name, uint32 flags, void **cookie)
{
    if (atomic_add(&pcm_open_count, 1) == 0) {
		return pcm_init();
    }
    return B_NO_ERROR;
}

static status_t pcm_close(void *cookie)
{
    return B_NO_ERROR;
}

static status_t pcm_free(void *cookie)
{
    if (atomic_add(&pcm_open_count, -1) == 1) {
		return pcm_uninit();
    }
    return B_NO_ERROR;
}

static status_t pcm_control(void *cookie, uint32 op, void *data, size_t length)
{
    switch (op) {
    case SOUND_GET_PARAMS:
		return pcm_get_params((sound_params *) data);
			
    case SOUND_SET_PARAMS:
		return pcm_set_params((sound_params *) data);
			
    case SOUND_SET_PLAYBACK_COMPLETION_SEM:
		return pcm_set_playback_sem(*((sem_id *) data));
			
    case SOUND_SET_CAPTURE_COMPLETION_SEM:
		return pcm_set_capture_sem(*((sem_id *) data));
			
    case SOUND_WRITE_BUFFER:
		return pcm_write_buffer((sound_buffer_header *) data);
			
    case SOUND_READ_BUFFER:
		return pcm_read_buffer((sound_buffer_header *) data);
			
    case SOUND_LOCK_FOR_DMA:
		return pcm_lock_for_dma();
    }
    return EINVAL;
}

static status_t pcm_read(void *cookie, off_t pos, void *data, size_t *length)
{
    return EINVAL;
}

static status_t pcm_write(void *cookie, off_t pos, const void *data, size_t *length)
{
    return EINVAL;
}


/********************************************************************
 * MIDI Input / Output Routines
 ********************************************************************/

static struct {
	generic_mpu401_module *module;
	void *driver;
} midi;

static void midi_callback(void)
{
	if (midi.module != 0 && midi.driver != 0)
		midi.module->interrupt_hook(midi.driver);
}

static void midi_interrupt_op(int32 op, void *card)
{
	if (op == B_MPU_401_ENABLE_CARD_INT) {
		hw_set_mpu401_callback(midi_callback)	;
	}
	else if (op == B_MPU_401_DISABLE_CARD_INT) {
		hw_set_mpu401_callback(0);
	}
}

static status_t midi_init(void)
{
	memset(&midi, 0, sizeof(midi));
	
	if (get_module(B_MPU_401_MODULE_NAME, (module_info **) &midi.module) != B_NO_ERROR)
		return ENOSYS;

	if (midi.module->create_device(hw_get_mpu401_port(), &midi.driver, 0, midi_interrupt_op, 0) != B_NO_ERROR) {
		put_module(B_MPU_401_MODULE_NAME);
		return B_ERROR;
	}
	
	return B_NO_ERROR;
}

static status_t midi_uninit(void)
{
	midi.module->delete_device(midi.driver);
	
	put_module(B_MPU_401_MODULE_NAME);
	
	return B_NO_ERROR;
}

static status_t midi_open(const char *name, uint32 flags, void **cookie)
{
	if (atomic_add(&midi_open_count, 1) == 0) {
		if (midi_init() != B_NO_ERROR) {
			atomic_add(&midi_open_count, -1);
			return B_ERROR;	
		}
	}
	return midi.module->open_hook(midi.driver, flags, cookie);
}

static status_t midi_close(void *cookie)
{
	return midi.module->close_hook(cookie);
}

static status_t midi_free(void *cookie)
{
	midi.module->free_hook(cookie);
	if (atomic_add(&midi_open_count, -1) == 1) {
		if (midi_uninit() != B_NO_ERROR)
			return B_ERROR;
	}
	return B_NO_ERROR;
}

static status_t midi_control(void *cookie, uint32 op, void *data, size_t len)
{
	return midi.module->control_hook(cookie, op, data, len);
}

static status_t midi_read(void *cookie, off_t pos, void *data, size_t *nread)
{
	return midi.module->read_hook(cookie, pos, data, nread);
}

static status_t midi_write(void *cookie, off_t pos, const void *data, size_t *nwritten)
{
	return midi.module->write_hook(cookie, pos, data, nwritten);
}


/********************************************************************
 * Device Driver Interface
 ********************************************************************/

int32 api_version = B_CUR_DRIVER_API_VERSION;

static const char *sb16_device_names[] = {
    "audio/old/sb16",
    "midi/sb16",
    0
};

static device_hooks sb16_device_hooks[] = {
    { pcm_open, pcm_close, pcm_free, pcm_control, pcm_read, pcm_write },
    { midi_open, midi_close, midi_free, midi_control, midi_read, midi_write }
};

status_t init_hardware(void)
{
    return hw_detect();
}

status_t init_driver(void)
{
    /* initialize the hardware */
    if (hw_init() != B_NO_ERROR)
		return ENODEV;
    
    /* reset the open counter */
    pcm_open_count = midi_open_count = 0;
    
    return B_NO_ERROR;
}

void uninit_driver(void)
{
    /* uninitialize the hardware */
    hw_done();
}

const char **publish_devices(void)
{
    /* return the list of device names */
    return sb16_device_names;
}

device_hooks *find_device(const char *name)
{
    int index;
    /* return the device hooks of the given device name */
    for (index = 0; sb16_device_names[index] != NULL; index++)
	if (strcmp(name, sb16_device_names[index]) == 0)
	    return &sb16_device_hooks[index];
    return NULL;
}
